# -*- coding: utf-8 -*-

"""
    frink.datastore
    ~~~~~~~~~~~~~
    Flask-Security Frink datastore classes.

"""

from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA

from flask.ext.security.datastore import Datastore, UserDatastore
from flask.ext.security.utils import get_identity_attributes

from .errors import NotUniqueError


class FrinkDatastore(Datastore):
    def put(self, model):
        model.save()
        return model

    def delete(self, model):
        model.delete()


class FrinkUserDatastore(FrinkDatastore, UserDatastore):
    """A RethinkDB / Schematics datastore implementation for Flask-Security
    """
    def __init__(self, db, user_model, role_model):
        FrinkDatastore.__init__(self, db)
        UserDatastore.__init__(self, user_model, role_model)

    def get_user(self, identifier):
        # Note: identifier here is probably an email address
        user = self.user_model.query.get(identifier)
        if user is not None:
            return user

        for attr in get_identity_attributes():
            column = getattr(self.user_model, attr)
            user = self.user_model.query.get_by(column=column.name, value=identifier)
            if user is not None:
                return user
            else:
                return None

    def find_user(self, **kwargs):
        print(cyan('find_user({})'.format(kwargs)))
        if kwargs.get('id', None) is not None:
            user = self.user_model.query.get(kwargs['id'])
            if user is not None:
                return user
        # If it's looking up by something else...
        user = self.user_model.query.first(**kwargs)
        if user is not None:
            return user

        return None

    def find_role(self, role):
        return self.role_model.query.first(name=role)

    def create_user(self, **kwargs):
        """Creates and returns a new user from the given parameters."""
        kwargs = self._prepare_create_user_args(**kwargs)
        user = self.user_model(kwargs)
        if self.get_user(user.email) is not None:
            raise NotUniqueError
        try:
            user.validate()
        except:
            raise
        else:
            user = self.put(user)
            return user
        return False



# -*- coding: utf-8 -*-

"""
    frink.errors
    ~~~~~~~~~~~~~
    Exception and Error classes
"""

import logging
from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA


def get_log(extra=None):
    m = "{}.{}".format(__name__, extra) if extra else __name__
    return logging.getLogger(m)

log = get_log()


class FrinkError(Exception):
    pass


class NotUniqueError(FrinkError):
    pass



# -*- coding: utf-8 -*-

"""
    frink.connection
    ~~~~~~~~~~~~~
    Setup of RethinkDB stuff.
"""

from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA

from flask import abort
from functools import wraps
from flask import current_app as app
from contextlib import contextmanager
import rethinkdb as r
from rethinkdb.errors import RqlRuntimeError, RqlDriverError
from .registry import model_registry


class rconnect(object):
    def __enter__(self):
        print(green('CONNECT ENTER'))
        try:
            self.conn = r.connect(
                host=app.config.get('RDB_HOST'),
                port=app.config.get('RDB_PORT'),
                db=app.config.get('RDB_DB')
            )
        except RqlDriverError:
            abort(503, "Database connection could be established.")
        else:
            return self.conn

    def __exit__(self, type, value, traceback):
        print(green('CONNECT EXIT'))
        self.conn.close()


class RethinkDB(object):

    def connection(self, app):
        return r.connect(host=app.config.get('RDB_HOST'), port=app.config.get('RDB_PORT'))

    def setup(self, app):
        connection = r.connect(host=app.config.get('RDB_HOST'), port=app.config.get('RDB_PORT'))
        try:
            r.db_create(app.config.get('RDB_DB')).run(connection)
            print('Database setup completed')
        except RqlRuntimeError:
            print('Database already exists.')
        finally:
            connection.close()

    def drop_all(self, app):
        conn = self.connection(app)
        return r.db_drop(app.config['RDB_DB']).run(conn)

    def init_app(self, app):

        print(magenta('RethinkSetup.init_app'))

        self.setup(app)

        # open connection before each request
        @app.before_request
        def before_request():
            print(green('CONNECT'))
            try:
                app.rdb_conn = r.connect(
                    host=app.config.get('RDB_HOST'),
                    port=app.config.get('RDB_PORT'),
                    db=app.config.get('RDB_DB')
                )
                print(green(app.rdb_conn))
            except RqlDriverError:
                abort(503, "Database connection could be established.")

        @app.teardown_request
        def teardown_request(exception):
            print(green('DISCONNECT'))
            try:
                app.rdb_conn.close()
            except AttributeError:
                pass

        # And finally...
        model_registry.init_app(app)



# -*- coding: utf-8 -*-

"""
    frink.base
    ~~~~~~~~~~~~~
    BaseModel class
"""

from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA

import datetime
from schematics.models import Model
from schematics.types.base import (
    StringType, BooleanType, DateTimeType, IntType, UUIDType
)

from .orm import ORMMeta
from .orm import InstanceLayerMixin


class BaseModel(Model, InstanceLayerMixin):

    # __metaclass__ = ORMMeta  # Do this in every model instead

    id = UUIDType()
    created_at = DateTimeType(default=datetime.datetime.now)
    updated_at = DateTimeType(default=datetime.datetime.now)

    def validate(self):
        print('validating')
        super(BaseModel, self).validate()

    def __repr__(self):
        if hasattr(self, 'email'):
            return u'<{}: {}>'.format(self.__class__.__name__, self.email)
        if hasattr(self, 'slug'):
            return u'<{}: {}>'.format(self.__class__.__name__, self.slug)
        if hasattr(self, 'name'):
            return u'<{}: {}>'.format(self.__class__.__name__, self.name)
        if hasattr(self, 'id'):
            return u'<{}: {}>'.format(self.__class__.__name__, self.id)
        return u'<{}: {} object>'.format(self.__class__.__name__, self.__class__.__name__)



# -*- coding: utf-8 -*-

"""
    frink.orm
    ~~~~~~~~~~~~~
    ORM type functionality
"""

from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA

import uuid
import logging
import rethinkdb as r
from inflection import tableize
from schematics.models import ModelMeta
from schematics.exceptions import ModelValidationError, ValidationError, ModelConversionError

from .errors import FrinkError, NotUniqueError
from .registry import model_registry
from .connection import rconnect


def get_log(extra=None):
    m = "{}.{}".format(__name__, extra) if extra else __name__
    return logging.getLogger(m)

log = get_log()


class InstanceLayerMixin(object):

    def save(self):
        """
            Save the current instance to the DB
        """
        with rconnect() as conn:
            try:
                self.validate()
            except ValidationError as e:
                print(red(e.messages))
                raise
            except ModelValidationError as e:
                print(red(e.messages))
                raise
            except ModelConversionError as e:
                print(red(e.messages))
                raise
            except Exception as e:
                print(red(e.messages))
                raise
            else:
                # If this is a new unsaved object, it'll likely have an
                # id of None, which RethinkDB won't like. So if it's None,
                # generate a UUID for it. If the save fails, we should re-set
                # it to None.
                if self.id is None:
                    self.id = str(uuid.uuid4())
                    log.info(self.id)

                try:
                    rv = r.db(self._db).table(self._table).insert(
                        self.to_primitive(),
                        conflict="replace"
                    ).run(conn)
                    # Returns something like this:
                    # {
                    #   u'errors': 0,
                    #   u'deleted': 0,
                    #   u'generated_keys': [u'dd8ad1bc-8609-4484-b6c4-ed96c72c03f2'],
                    #   u'unchanged': 0,
                    #   u'skipped': 0,
                    #   u'replaced': 0,
                    #   u'inserted': 1
                    # }
                    log.info(green(rv))
                except Exception as e:
                    log.warn(red(e))
                    self.id = None
                else:
                    return self

    def delete(self):
        """
            Delete the current instance from the DB.
        """
        with rconnect() as conn:
            # Can't delete an object without an ID.
            if self.id is None:
                raise FrinkError
            else:
                if isinstance(self.id, uuid.UUID):
                    self.id = str(self.id)

                try:
                    rv = r.db(
                        self._db
                    ).table(
                        self._table
                    ).get(
                        self.id
                    ).delete().run(conn)
                    log.info(green(rv))
                except Exception as e:
                    log.warn(red(e))
                    raise
                else:
                    return True
        return False


class ORMMeta(ModelMeta):

    instance = None

    def __new__(cls, name, bases, dct):
        if name != "BaseModel":
            # We have a model class
            cls._table = tableize(name)
            super_new = super(ORMMeta, cls).__new__
            new_class = super_new(cls, name, bases, dct)
            cls._model = new_class
            cls._orm = ORMLayer(cls._table, cls._model)
            setattr(new_class, 'query', cls._orm)
            setattr(new_class, '_table', cls._table)
            # register the model
            model_registry.add(name, new_class, cls)

        return new_class


class ORMLayer(object):

    def __init__(self, table, model):
        self._model = model
        self._table = table
        print('query init {}'.format(self._table))

    def get(self, id):
        """
            Get a single instance by pk id.

            :param id: The UUID of the instance you want to retrieve.

        """
        with rconnect() as conn:
            if id is None:
                raise ValueError

            if isinstance(id, uuid.UUID):
                id = str(id)

            if type(id) != str and type(id) != unicode:
                raise ValueError

            try:
                rv = self._base().get(id).run(conn)
            except Exception as e:
                print(red(e))
                raise
            if rv is not None:
                return self._model(rv)

            return None

    def filter(self, order_by=None, limit=0, **kwargs):
        """
            Fetch a list of instances.

            :param order_by: column on which to order the results. \
            To change the sort, prepend with < or >.
            :param limit: How many rows to fetch.
            :param kwargs: keyword args on which to filter, column=value

        """
        with rconnect() as conn:
            print(magenta(kwargs))
            if len(kwargs) == 0:
                raise ValueError

            try:
                query = self._base()
                query = query.filter(kwargs)

                if order_by is not None:
                    query = self._order_by(query, order_by)

                if limit > 0:
                    query = self._limit(query, limit)

                print(magenta(query))
                rv = query.run(conn)

            except Exception as e:
                print(red(e))
                raise
            else:
                data = [self._model(_) for _ in rv]
                return data

    def first(self, order_by=None, **kwargs):
        """
            Fetch the first item that matches your filter.

            :param order_by: column on which to order the results. \
            To change the sort, prepend with < or >.
            :param kwargs: keyword args on which to filter, column=value

        """
        with rconnect() as conn:
            if len(kwargs) == 0:
                raise ValueError

            try:
                query = self._base()
                if order_by is not None:
                    query = self._order_by(query, order_by)

                # NOTE: Always filter before limiting
                query = query.filter(kwargs)
                query = self._limit(query, 1)
                rv = query.run(conn)
            except Exception as e:
                print(red(e))
                raise
            else:
                data = [self._model(_) for _ in rv]
                try:
                    return data[0]
                except IndexError:
                    return None

    def find_by(self, column=None, value=None, order_by=None, limit=0):
        """
            Find all items that matches your a column/value.

            :param column: column to search.
            :param value: value to look for in `column`.
            :param limit: How many rows to fetch.
            :param order_by: column on which to order the results. \
            To change the sort, prepend with < or >.

        """
        with rconnect() as conn:
            if column is None or value is None:
                raise ValueError("You need to supply both a column and a value")

            try:
                query = self._base()
                if order_by is not None:
                    query = self._order_by(query, order_by)
                if limit > 0:
                    query = self._limit(query, limit)

                rv = query.filter({column: value}).run(conn)
            except Exception as e:
                print(red(e))
                raise
            else:
                data = [self._model(_) for _ in rv]
                return data

    def get_by(self, column=None, value=None):
        """
            Find one item that matches your a column/value. Raises `NotUniqueError` if
            it finds more than one.

            :param column: column to search.
            :param value: value to look for in `column`.
            :param limit: How many rows to fetch.
            :param order_by: column on which to order the results. \
            To change the sort, prepend with < or >.

        """
        with rconnect() as conn:
            if column is None or value is None:
                raise ValueError("You need to supply both a column and a value")

            try:
                query = self._base()
                rv = query.filter({column: value}).run(conn)
            except Exception as e:
                print(red(e))
                raise
            else:
                data = [self._model(_) for _ in rv]
                if len(data) > 1:
                    raise NotUniqueError('Found more than one object matching the query')
                else:
                    try:
                        return data[0]
                    except IndexError:
                        return None

    def all(self, order_by=None, limit=0):
        """
            Fetch all items.

            :param limit: How many rows to fetch.
            :param order_by: column on which to order the results. \
            To change the sort, prepend with < or >.

        """
        with rconnect() as conn:
            try:
                query = self._base()
                if order_by is not None:
                    query = self._order_by(query, order_by)
                if limit > 0:
                    query = self._limit(query, limit)

                rv = query.run(conn)
            except Exception as e:
                print(red(e))
                raise
            else:
                data = [self._model(_) for _ in rv]
                return data

    def _limit(self, query, limit):
        with rconnect() as conn:
            try:
                rv = query.limit(limit)
            except Exception as e:
                print(red(e))
                raise
            else:
                return rv

    def _order_by(self, query, column):
        with rconnect() as conn:
            if column.startswith('>'):
                index = r.desc(column[1:])
            elif column.startswith('<'):
                index = r.asc(column[1:])
            else:
                index = column

            try:
                rv = query.order_by(index)
            except Exception as e:
                print(red(e))
                raise
            else:
                return rv

    def _base(self):
        return r.db(self._db).table(self._table)



import setuptools


def get_requirements(suffix=''):
    with open('requirements%s.txt' % suffix) as f:
        rv = f.read().splitlines()
    return rv


setuptools.setup(
    name="frink",
    version="0.0.3",
    url="https://github.com/hactar-is/frink",

    author="Hactar",
    author_email="systems@hactar.is",

    description="Basic ORM-like functionality for a RethinkDB datastore on top of Schematics.",
    long_description="""Basic ORM-like functionality for a RethinkDB datastore on top of
    Schematics. Includes datastores for use with Flask-Security.""",

    packages=setuptools.find_packages(),

    install_requires=get_requirements(),

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7'
    ],
)



# -*- coding: utf-8 -*-

"""
    frink.registry
    ~~~~~~~~~~~~~
    Keep a record of all the models registered.
"""

from fabric.colors import green, red, blue, cyan, magenta, yellow  # NOQA
import rethinkdb as r
from rethinkdb.errors import ReqlOpFailedError


class ModelRegistry(object):
    def __init__(self):
        self._models = {}
        self._tables = []
        self._initialsed = False
        self._app = None

    def add(self, name, model, meta):
        print(green('Registering {} (initialised: {})'.format(
            name, self._initialsed
        )))

        if name in self._models:
            return
        if type(model) != meta:
            raise ValueError('{} is not of type({})'.format(model, meta))
        self._models[name] = model
        if self._initialsed is True:
            # This model is appearing after init_app has already
            # run, so we need to _init_model it.
            self._init_model(name, model)

    def init_app(self, app):
        print(green('registry init_app'))
        self._app = app
        for name, model in self._models.items():
            self._init_model(name, model)
        self._initialsed = True

    def _init_model(self, name, model):
        model._model = model
        setattr(model, '_db', self._app.config.get('RDB_DB', 'test'))
        setattr(model.query, '_db', self._app.config.get('RDB_DB', 'test'))
        if name not in self._tables:
            try:
                print(green('create {} table for {}'.format(model._table, name)))
                self._conn = r.connect(
                    host=self._app.config.get('RDB_HOST'),
                    port=self._app.config.get('RDB_PORT'),
                    db=self._app.config.get('RDB_DB')
                )
                r.db(model._db).table_create(model._table).run(self._conn)
                self._conn.close()
            except ReqlOpFailedError as e:
                print(yellow('{} table probably already exists'.format(model._table)))
            except Exception as e:
                print(red(e))
                raise
            else:
                self._tables.append(name)


model_registry = ModelRegistry()
